# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import math

import soya, soya.cursor as cursor
import py2play.action as action

from slune.character import *
from slune.level     import Level

def Debug(character):
  yield Action(ACTION_DEBUG, 0.0, -1.4)
  yield Action(ACTION_DEBUG, 0.0, -1.4)
  yield Action(ACTION_DEBUG, 0.0, -1.4)
  yield Action(ACTION_DEBUG, 0.0, -1.4)
  
def Wall_climber(character):
  yield Action(ACTION_JUMP, character.default_mouse_x(), -1.0)
  while character.on_ground:
    yield Action(ACTION_ROTUP, character.default_mouse_x(), -1.0)
  i = 0
  while (character.force.y > 0.0) or (i < 0.0 * 7.0 / character.rotup_speed):
    i += 1
    yield Action(ACTION_ROTUP, character.default_mouse_x(), -1.0)
  i = 0
  while (not character.on_ground) and (i < 25.0 * 7.0 / character.rotup_speed):
    i += 1
    yield Action(ACTION_ROTDOWN, character.default_mouse_x(), -1.0)

def Dodge(character):
  yield Action(ACTION_JUMP, 0.0, 0.0)
  
  for j in range(30):
    yield Action(ACTION_ROTUP, 0.0, 0.0)
    
  for j in range(30):
    yield Action(ACTION_WAIT, 0.0, 0.0)
    
def nearest_character(character):
  best      = None
  best_dist = 10000.0
  for level in character.get_root():
    if isinstance(level, Level):
      for charac in level.characters:
        if not charac is character:
          dist = character.distance_to(charac)
          if dist < best_dist:
            best      = charac
            best_dist = dist
            
  return best
  
def Goto_next_flag(character):
  if not hasattr(character, "next_flag_name"): return
  flag = character.parent[character.next_flag_name]
  if flag: yield Runto(character, flag, character.radius + flag.radius)
  
def Push_nearest_character(character):
  nearest = nearest_character(character)
  if nearest: yield Goto(character, nearest, character.radius + nearest.radius)
  
def Look_at_nearest_character(character):
  nearest = nearest_character(character)
  if nearest: yield Look_at(character, nearest)
  
def Goto(character, target, min_distance = 0.7):
  mouse_x = 0.0
  while character.distance_to(target) > min_distance:
    fx, fy, fz = character.transform_point(target.x, target.y, target.z, target.parent)
    
    if fz < 0.0:
      mouse_x = 0.8 * mouse_x + 0.2 * max(-1.0, min(1.0, -4.0 * math.atan(fx / fz)))
      yield Action(ACTION_WAIT, mouse_x, max(-1.0, min(0.0, fz + abs(fx)) / 5.0 + 1.0))
      #yield Action(ACTION_WAIT,
      #             mouse_x = min(1.0, max(-1.0, 0.23161940934461928 * fx)),
      #             mouse_y = min(1.0, max(-1.0, 0.54152636319441949 * abs(fx) + 0.057433463934462713 * fy + 1.5669823329989161 * fz)),
      #             ) 
    else:
      yield Action(ACTION_WAIT, cmp(fx, 0.0), 1.4)
      
def Runto(character, target, min_distance = 0.7):
  mouse_x = 0.0
  while character.distance_to(target) > min_distance:
    fx, fy, fz = character.transform_point(target.x, target.y, target.z, target.parent)
    
    if fz < 0.0:
      mouse_x = 0.8 * mouse_x + 0.2 * max(-1.0, min(1.0, -4.0 * math.atan(fx / fz)))
      yield Action(ACTION_WAIT, mouse_x, min(1.0, -1.0 + 2.0 * abs(mouse_x)))
    else:
      yield Action(ACTION_WAIT, cmp(fx, 0.0), 1.4)

Runto = Goto

#  while character.distance_to(target) > min_distance:
#    x, y, z = character.transform_point(target.x, target.y, target.z, target.parent)
#    
#    if z < 0.0:
#      yield Action(ACTION_WAIT, max(-1.0, min(1.0, -2.0 * x / z)), -1.0)
#    else:
#      yield Action(ACTION_WAIT, cmp(x, 0.0), 1.0)
      
def Look_at(character, target):
  
  while 1:
    x, y, z = character.transform_point(target.x, target.y, target.z, target.parent)
    
    if abs(x / z) < 0.2: break
    
    if z < 0.0:
      yield Action(ACTION_WAIT, cmp(x, 0.0), 1.4) 
    else:
      yield Action(ACTION_WAIT, cmp(x, 0.0), 1.4)
      
def STAY_HERE(): return Action(ACTION_WAIT, 0.0, 1.4)

class Wait:
  def next(self):
    return STAY_HERE()
WAIT = Wait()


class StackControler:
  def __init__(self, controlers  = None):
    self.controlers = controlers or []
    self.uncancelable_controler = None
    
  def animate(self, controler):
    self.controlers.append(controler)
    if not self.uncancelable_controler in self.controlers:
      self.uncancelable_controler = controler
      
  def append(self, controler):
    if not self.uncancelable_controler in self.controlers:
      self.controlers.append(controler)
      
  def next(self):
    try:
      c = self.controlers[-1]
      r = c.next() # calling c.net() may add a new controler ! => c may no longer be self.controlers[-1]
    except StopIteration:
      self.controlers.remove(c)
      if self.uncancelable_controler is c: self.uncancelable_controler = None
      return self.next()
    
    if not isinstance(r, Action):
      self.controlers.append(r)
      return self.next()
    
    return r

# A controler for phantom character, needed for them to play video sequences.
class PhantomControler(StackControler):
  def next(self):
    if self.controlers: return StackControler.next(self)


class KeyboardMouseControler(StackControler):
  key_controlers = {}
  
  def __init__(self, character = None):
    StackControler.__init__(self)
    
    self.d_mouse_x     = self.d_mouse_y = 0.0
    self.mouse_x       = 0.0
    self.mouse_y       = 1.4
    self.character     = character
    self.enabled       = 1
    self.keyboard_mode = 0.0
    
  def next(self):
    character = self.character
    a = character.player.default_action
    
    if (character.player.default_action == ACTION_ROTUP) or (character.player.default_action == ACTION_ROTDOWN):
      if character.on_ground:
        character.player.default_action = a = ACTION_WAIT
        
    for event in cursor.coalesce_motion_event(soya.process_event()):
      if   event[0] == soya.MOUSEMOTION:
        self.keyboard_mode = 0
        camera = character.get_root().camera
        self.mouse_x = ((float(event[1]) / camera.get_screen_width ()) - 0.5) * 2.0
        self.mouse_y = ((float(event[2]) / camera.get_screen_height()) - 0.5) * 2.0
        
      elif self.enabled:
        if self.controlers and (event[0] != soya.MOUSEBUTTONUP) and (not self.uncancelable_controler in self.controlers):
          self.controlers *= 0
          
        if event[0] == soya.MOUSEBUTTONDOWN:
          if   event[1] == 1:
            if character.on_ground: a = ACTION_JUMP
            else:             character.player.default_action = a = ACTION_ROTUP
          elif event[1] == 2:
            if character.on_ground:
              character.player.controler.append(Wall_climber(character))
              return
            else: character.player.default_action = a = ACTION_ROTDOWN
          elif event[1] == 3: character.player.default_action = a = ACTION_ROTLAT
          elif event[1] == 4:
            if character.player.default_action == ACTION_ROTUP  : character.player.default_action = a = action.ACTION_WAIT
            else:                                                 character.player.default_action = a = ACTION_ROTDOWN
          elif event[1] == 5:
            if character.player.default_action == ACTION_ROTDOWN: character.player.default_action = a = action.ACTION_WAIT
            else:                                                 character.player.default_action = a = ACTION_ROTUP
            
        elif event[0] == soya.MOUSEBUTTONUP:
          if   event[1] == 1: character.player.default_action = a = ACTION_WAIT
          elif event[1] == 3: character.player.default_action = a = ACTION_WAIT
          
        elif event[0] == soya.KEYDOWN:
          if   (event[1] == soya.K_LCTRL) or (event[1] == soya.K_RCTRL):
            if character.on_ground: a = ACTION_JUMP
            else: character.player.default_action = a = ACTION_ROTUP
            
          elif (event[1] == soya.K_LSHIFT) or (event[1] == soya.K_RSHIFT):
            if character.on_ground:
              character.player.controler.append(Wall_climber(character))
              return
            else: character.player.default_action = a = ACTION_ROTDOWN
            
          elif (event[1] == soya.K_LALT) or (event[1] == soya.K_RALT):
            character.player.default_action = a = ACTION_ROTLAT
            
          elif event[1] == soya.K_DOWN:  self.keyboard_mode = 1; self.d_mouse_y =  0.15
          elif event[1] == soya.K_UP:    self.keyboard_mode = 1; self.d_mouse_y = -0.15
          elif event[1] == soya.K_LEFT:  self.keyboard_mode = 1; self.d_mouse_x = -0.15
          elif event[1] == soya.K_RIGHT: self.keyboard_mode = 1; self.d_mouse_x =  0.15
          
          elif (event[1] == soya.K_q) or (event[1] == soya.K_ESCAPE):
            from py2play.idler import IDLER
            IDLER.end_game()
            
          #elif event[1] == 102: a = ACTION_TURBO
          elif event[1] == soya.K_d:
            #a = ACTION_DEBUG
            character.player.controler.append(Debug(character))
            return
          
          elif event[1] ==  97:
            import time
            time.sleep(1.0)
            
          else:
            print event[1]
            
            Controler = self.key_controlers.get(event[1])
            if Controler:
              character.player.controler.append(Controler(character))
              
        elif event[0] == soya.KEYUP:
          if   event[1] == soya.K_DOWN:
            if self.d_mouse_y > 0.0: self.d_mouse_y = 0.0
          elif event[1] == soya.K_UP:
            if self.d_mouse_y < 0.0: self.d_mouse_y = 0.0
          elif event[1] == soya.K_LEFT:
            if self.d_mouse_x < 0.0: self.d_mouse_x = 0.0
          elif event[1] == soya.K_RIGHT:
            if self.d_mouse_x > 0.0: self.d_mouse_x = 0.0
            
          elif (event[1] == soya.K_LALT) or (event[1] == soya.K_RALT):
            character.player.default_action = a = ACTION_WAIT
            
      elif (event[0] == soya.MOUSEBUTTONDOWN) or (event[0] == soya.KEYDOWN):
        return Action(ACTION_STARTGAME, 0.0, 1.4)
        
    from py2play.idler import IDLER
    if self.controlers:
      if self.uncancelable_controler and not IDLER.blackbands_visible: IDLER.show_blackbands()
      
      return StackControler.next(self)
    
    else:
      if IDLER.blackbands_visible: IDLER.hide_blackbands()

      if self.keyboard_mode:
        self.mouse_x *= 0.9
        if self.mouse_y < 0.0: self.mouse_y *= 0.9
        
        if self.d_mouse_x and (-1.0 < self.mouse_x + self.d_mouse_x < 1.0):
          self.mouse_x += self.d_mouse_x
          
        if self.d_mouse_y and (-1.0 < self.mouse_y + self.d_mouse_y < 1.0):
          self.mouse_y += self.d_mouse_y
        
        
      return Action(a, self.mouse_x, self.mouse_y)
    
    
    
KeyboardMouseControler.key_controlers[soya.K_z] = Push_nearest_character #z
KeyboardMouseControler.key_controlers[soya.K_e] = Wall_climber #e
KeyboardMouseControler.key_controlers[soya.K_f] = Goto_next_flag #f
KeyboardMouseControler.key_controlers[soya.K_g] = Dodge #g



          
