# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya, _soya, soya.soya3d as soya3d, soya.model as model, soya.particle as particle, soya.game.level as game_level, soya.widget as widget, soya.ray as ray
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
from soya.math3d import Point, Vector
import slune.character as slune_character
import slune.controler
import slune.globdef as globdef, slune.sound as sound

def init_race_game(level):
  """Hides all flags except the first."""
  level.ranks = []
  
  level.flags = [] # Indexes all flags
  
  for item in level:
    if isinstance(item, Flag):
      item.visible = (item.name == "flag_0")
      level.flags.append(item)
      
  level.flags.sort(lambda f1, f2: cmp(f1.flag_id, f2.flag_id))
  
def start_race_game(level):
  """Inits time property and shows the race time label."""
  CheckEveryoneIsKilled(level)
  
  for character in level.characters:
    character.next_flag_id = 0
    character.nb_laps   = 0
    character.lap_start = time.time()
    character.best_lap  = 9999.0
    if character.played and character.player.active:
      from py2play.idler import IDLER
      IDLER.no_blackbands_group.add(LapsLabel(character))
      IDLER.no_blackbands_group.add(MiniMap  (level))
      
def end_race_game(level):
  """Hides the race time label."""
  for widget in soya.root_widget.children:
    if isinstance(widget, LapsLabel):
      soya.root_widget.remove(widget)
      break
    
class CheckEveryoneIsKilled(soya3d.Volume):
  def begin_round(self):
    from py2play.player import CURRENT_PLAYER
    
    if (len(self.parent.characters) == 1) and (self.parent.characters[0] is CURRENT_PLAYER.character):
      from py2play.idler import IDLER
      
      self.parent.ranks.append(CURRENT_PLAYER.character)
      if len(self.parent.ranks) == 1:
        IDLER.level_completed(CURRENT_PLAYER.character, 1, message = _("__race_winner__"))
      else:
        IDLER.level_completed(CURRENT_PLAYER.character, 1, message = _("__race_looser__") % len(self.parent.ranks))
        
      self.parent.remove(self)
  
  
class Flag(soya3d.Volume):
  radius = 3.5
  
  def __init__(self, parent = None):
    if parent:
      parent.max_laps = 2
      
      nb = 0
      for item in parent:
        if isinstance(item, Flag): nb += 1
        
      soya3d.Volume.__init__(self, parent, model.Shape.get("flag"))
      
      self.flag_id = nb
      
    else:
      soya3d.Volume.__init__(self, parent, model.Shape.get("flag"))
      
  def get_flag_id(self):
    try: return int(self.name[5:])
    except ValueError: return 0
  def set_flag_id(self, id): self.name = "flag_%s" % id
  flag_id = property(get_flag_id, set_flag_id)
  
  def begin_round(self):
    from py2play.idler import IDLER
    if IDLER and IDLER.waiting_until_ready: return
    
    level = self.parent
    while not isinstance(level, py2p_level.Level):
      if level is None: return
      level = level.parent
      
    for character in level.characters:
      if not hasattr(character, "next_flag_id"):
        init_race_opponent(character)
        
      if (character.next_flag_id == self.flag_id) and (character.distance_to(self) < self.radius + character.radius):
        if character.next_flag_id + 1 < len(self.parent.flags):
          next_flag_id = character.next_flag_id + 1
          next_flag = self.parent.flags[next_flag_id]
        else:
          next_flag_id = 0
          next_flag = self.parent.flags[0]
          character.nb_laps += 1
          now = time.time()
          this_lap = now - character.lap_start
          if this_lap < character.best_lap: character.best_lap = this_lap
          character.lap_start = now
          
          if character.nb_laps >= globdef.RACE_NB_LAPS:
            firework = particle.Firework(self.parent, nb_particles = 6)
            firework.move(self)
            firework.generator.initial_speed = 0.5
            firework.generator.acceleration.y = -0.05
            
            character.next_flag_id = 0
            
            level.ranks.append(character)
            
            if len(level.ranks) == 1: # Winner
              IDLER.level_completed(character, 1, _("__race_winner__"))
            else:
              IDLER.level_completed(character, 0, _("__race_looser__") % len(level.ranks))
              
            from py2play.player import CURRENT_PLAYER
            if not character is CURRENT_PLAYER.character:
              IDLER.message(_("__race_arrival__") % (character.displayname(), len(level.ranks)))
              
            if not character.played:
              character.level.remove_character(character)
            return
          
        character.next_flag_id = next_flag_id
        
        sound.play("flag-1.wav", character, character.speed)
        
        if character.played and character.player.active:
          self.visible = 0
          next_flag.visible = 1
          firework = particle.Firework(self.parent, nb_particles = 1)
          firework.move(self)
          firework.look_at(Vector(character, 0.0, 0.0, -1.0))
          firework.turn_vertical(-50.0)
          firework.removable = 1
          firework.generator.initial_speed = 0.5
          firework.generator.acceleration.y = -0.05
          
  def advance_time(self, proportion):
    if self.visible: self.rotate_lateral(proportion * 5.0)

import os.path

def init_race_opponent(character):
  character.next_flag_id = 0
  character.nb_laps   = 0
  character.lap_start = time.time()
  character.best_lap  = 9999.0
  if character.played and character.player.active:
    soya.root_widget.add(LapsLabel(character))
    
    for flag in character.level.flags: flag.visible = 0
    character.level.flags[0].visible = 1
        

PERSO2COLOR = {
  "tux" : (0.2, 0.2, 0.2, 1.0),
  }

class MiniMap(widget.Widget):
  def __init__(self, level):
    self.level = level
    
    self.top = 10
    self.width  = 200
    self.height = 200
    self.left = soya.root_widget.width - self.width - 40
    
    flags = filter(lambda x: isinstance(x, Flag), level)
    flags.sort(lambda f1, f2: cmp(f1.flag_id, f2.flag_id))
    min_x = min_y =  10e1000
    max_x = max_y = -10e1000
    for flag in flags:
      if flag.x < min_x: min_x = flag.x
      if flag.x > max_x: max_x = flag.x
      if flag.z < min_y: min_y = flag.z
      if flag.z > max_y: max_y = flag.z
    f_x = self.width  / (max_x - min_x)
    f_y = self.height / (max_y - min_y)
    if f_x < f_y:
      f = f_x
      left = self.left
    else:
      f = f_y
      left = self.left + ((max_y - min_y) - (max_x - min_x)) * f
    top = self.top
    
    self.flags_pos = []
    self._calllist = _soya.glGenList()
    _soya.glNewList(self._calllist)
    _soya.glColor4f(1.0, 1.0, 1.0, 1.0)
    _soya.glBegin(_soya.GL_LINE_LOOP)
    for flag in flags:
      self.flags_pos.append((
        left + f * (flag.x - min_x),
        top  + f * (flag.z - min_y),
        ))
      _soya.glVertex2f(*self.flags_pos[-1])
    _soya.glEnd()
    x, y = self.flags_pos[-1]
    _soya.glBegin(_soya.GL_QUADS)
    _soya.glVertex2f(x - 4.0, y - 4.0)
    _soya.glVertex2f(x - 4.0, y + 4.0)
    _soya.glVertex2f(x + 4.0, y + 4.0)
    _soya.glVertex2f(x + 4.0, y - 4.0)
    _soya.glEnd()
    
    _soya.glEndList()
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.left = parent_width - self.width - 40
    self.screen_width  = parent_width
    self.screen_height = parent_height
    
  def render(self):
    from py2play.player import CURRENT_PLAYER
    from py2play.idler import IDLER
    ca = IDLER.camera
    x, y, z = ca.transform_point(0.0, 0.0, 0.0, self.level.flags[CURRENT_PLAYER.character.next_flag_id])
    angle = math.atan(x / z)
    if z < 0.0:
      vx = -math.sin(angle)
      vy = -math.cos(angle)
    else:
      vx = math.sin(angle)
      vy = math.cos(angle)
    
    r = self.screen_width / self.screen_height
    if (self.screen_width / self.screen_height) > abs(vx / vy):
      if vy < 0.0:
        my = -self.screen_height / 2.0
        mx = vx * (my / vy) * r
      else:
        my = self.screen_height / 2.0
        mx = vx * (my / vy) * r
    else:
      if vx < 0.0:
        mx = -self.screen_width / 2.0
        my = vy * (mx / vx) / r
      else:
        mx = self.screen_width / 2.0
        my = vy * (mx / vx) / r
        
    x = mx + self.screen_width  / 2.0
    y = my + self.screen_height / 2.0
    _soya.glColor4f(1.0, 1.0, 1.0, 1.0)
    _soya.glBegin(_soya.GL_TRIANGLES)
    _soya.glVertex2f(x, y)
    _soya.glVertex2f(x + (+ my - mx) * 0.06, y + (- mx - my) * 0.06)
    _soya.glVertex2f(x + (- my - mx) * 0.06, y + (+ mx - my) * 0.06)
    _soya.glEnd()
    
    _soya.glCallList(self._calllist)
    for character in self.level.characters:
      if hasattr(character, "next_flag_id"):
        prev = self.level.flags[character.next_flag_id - 1]
        next = self.level.flags[character.next_flag_id]

        #if (prev >> next).dot_product(prev >> character) < 0.0:
        if (next.x - prev.x) * (character.x - prev.x) + (next.y - prev.y) * (character.y - prev.y) + (next.z - prev.z) * (character.z - prev.z) < 0.0:
          next = prev
          prev = self.level.flags[character.next_flag_id - 2]
          
        d1 = character.distance_to(prev)
        d2 = character.distance_to(next)
        d = d1 / (d1 + d2)

        prev_x = self.flags_pos[prev.flag_id][0]
        prev_z = self.flags_pos[prev.flag_id][1]
        x = prev_x + d * (self.flags_pos[next.flag_id][0] - prev_x)
        y = prev_z + d * (self.flags_pos[next.flag_id][1] - prev_z)
        
        if   character.perso_name == "tux":
          material = model.Material.get("tux1"); material.activate()
          _soya.glBegin(_soya.GL_QUADS)
          _soya.glTexCoord2f(0.6, 0.0); _soya.glVertex2f(x - 10.0, y - 10.0)
          _soya.glTexCoord2f(0.6, 1.0); _soya.glVertex2f(x - 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0, 1.0); _soya.glVertex2f(x + 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0, 0.0); _soya.glVertex2f(x + 10.0, y - 10.0)
          _soya.glEnd()
        elif character.perso_name == "gnu":
          material = model.Material.get("gnu"); material.activate()
          _soya.glBegin(_soya.GL_QUADS)
          _soya.glTexCoord2f(0.52, 0.4); _soya.glVertex2f(x - 10.0, y - 10.0)
          _soya.glTexCoord2f(0.52, 1.0); _soya.glVertex2f(x - 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0 , 1.0); _soya.glVertex2f(x + 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0 , 0.4); _soya.glVertex2f(x + 10.0, y - 10.0)
          _soya.glEnd()
        elif character.perso_name == "shark":
          material = model.Material.get("shark1"); material.activate()
          _soya.glBegin(_soya.GL_QUADS)
          _soya.glTexCoord2f(0.46, 0.48); _soya.glVertex2f(x - 10.0, y - 10.0)
          _soya.glTexCoord2f(0.46, 1.0 ); _soya.glVertex2f(x - 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0 , 1.0 ); _soya.glVertex2f(x + 10.0, y + 10.0)
          _soya.glTexCoord2f(1.0 , 0.48); _soya.glVertex2f(x + 10.0, y - 10.0)
          _soya.glEnd()
        material.inactivate()
        
        
class LapsLabel(widget.OptimizedLabel):
  def __init__(self, character):
    self.character = character
    self.format = str(_("__racelapsformat__"))
    widget.OptimizedLabel.__init__(self)
    
    self.left = 10
    self.top = soya.root_widget.height - 33 #100
    self.width  = 400
    self.height = 100
    self.old_nb_laps = -1
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.top = soya.root_widget.height - 33 #100
    
  def render(self):
    if self.old_nb_laps != self.character.nb_laps:
      self.text = self.format % (self.character.nb_laps)
      self.old_nb_laps = self.character.nb_laps
      #self.text = self.format % (self.character.nb_laps, self.character.best_lap, time.time() - self.character.lap_start)
    widget.OptimizedLabel.render(self)
    

CHROMOSOMS = {
  ("tux", 0): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.2 , 80.0, 30.0, 0.5],
  ("tux", 1): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.55, 30.0, 50.0, 0.5],
  ("tux", 2): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 1.0 , 20.0, 50.0, 0.6],
  
  ("gnu", 0): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.1, 50.0, 50.0, 0.7],
  ("gnu", 1): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.5 , 15.0, 80.0, 0.8],
  ("gnu", 2): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 1.0 , 10.0, 99.0, 0.8],
  
  ("shark", 0): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.1 , 50.0, 50.0, 0.4],
  ("shark", 1): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.5 , 15.0, 50.0, 0.3],
  ("shark", 2): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 1.0 , 10.0, 60.0, 0.2],
  }

class Opponent(slune_character.NonPlayerCharacter, slune_character.Competitor):
  def __init__(self, parent = None, chromosom = None):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    
    self.check_ground_vector = Vector(self, 0.0, -1.0, -2.0)
    
    # Arbitrary not-so-bad values
    #self.chromosom = chromosom or [0.3,  0.0, 0.0, 1.0, 100.0, 1.0, 0.1, 3.0]
    #self.chromosom = chromosom or [0.3, -0.5, 0.0, 1.0, 100.0, 1.0, 0.1, 3.0]
    #self.chromosom = chromosom or [0.3, -0.5, 0.0, 1.0, 100.0, 1.0, 0.1, 3.0]
    
    # Values obtained by genetic optimization (see race_algogen.py)
    self.chromosom = chromosom
    #self.chromosom = chromosom or [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322]
    #self.chromosom = chromosom or [0.46522031443857115, -0.14001113793863912, 0.025261052766113602, 0.6150540323882927, 5.9620468661604544, 0.94158933248522936, 0.079879170490081183, 3.3103171301896737]
    
  def mutate(self, value):
    if self.random.random() < 0.8: return value
    return value + (self.random.random() - 0.5) * value
  
  def init(self):
    if not self.perso_name:
      persos = list(globdef.CHARACTERS)
      self.random.shuffle(persos)
      for perso in persos:
        perso = perso.lower()
        for c in self.level:
          if (hasattr(c, "perso_name")) and (c.perso_name == perso): break
        else:
          self.perso_name = perso
          break
      else: self.perso_name = persos[0].lower()
    self.set_perso(self.perso_name)
    self.set_vehicle(self.random.choice(globdef.VEHICLES.values()))
    
  def IAControler(character):
    c = character.chromosom = character.chromosom or CHROMOSOMS[character.perso_name, character.level.difficulty]
    max_speed = c[8] # Do NOT mutate c[8] / maxspeed !!!
    c = map(character.mutate, c)
    c[8] = max_speed
    
    jumping = 0
    
    mouse_x = mouse_y = 0.0
    
    while not hasattr(character, "next_flag_id"):
      # Not yet fully initialized => wait !
      yield slune_character.Action(slune_character.ACTION_WAIT, 0.0, 1.4)
      
    character.speed.y = 100.0 # Avoid null speed (null speed => jumping). The speed will be overrided later.
    
    while 1:
      for i in range(c[9]):
        next_flag = character.level.flags[character.next_flag_id]
        fx, fy, fz = character.transform_point(next_flag.x, next_flag.y, next_flag.z, next_flag.parent)
        
        if jumping:
          if character.on_ground: jumping = 0
          no_ground = 0 # It is normal that there is no ground !
        else:
          #no_ground = not character.level.raypick_b(character, character.check_ground_vector, 4.0, 3)
          no_ground = character.level.raypick(character, character.check_ground_vector, 4.0, 3) is None
          
        if   (fy > 3.5) and (abs(fx) < 2.0) and (-10.0 < fz < 0.0):
          yield slune.controler.Wall_climber(character)
          jumping = 1
        elif (c[5] * fy + c[6] * fz + c[7] * no_ground > 2.0) or (character.speed.length() < 0.001):
          action = slune_character.ACTION_JUMP
          jumping = 1
        else:
          action = slune_character.ACTION_WAIT
          
        mouse_x = min(1.0, max(-1.0, c[0] * fx))
        mouse_y = min(max_speed, max(-max_speed, c[1] * abs(fx) + c[2] * fy + c[3] * fz + c[4] * no_ground))
        
        yield slune_character.Action(action, mouse_x, mouse_y)

      # Check boosting
      p = filter(lambda p: p.played, character.level.characters)
      if p:
        p = character.random.choice(p)
        delta = (character.nb_laps * len(character.level.flags) + character.next_flag_id) - (p.nb_laps * len(p.level.flags) + p.next_flag_id)
        if   (delta >  1.0) and p.level.difficulty < 2: max_speed = max_speed * 0.9
        elif (delta < -1.0) and p.level.difficulty > 0: max_speed = 1.1
        elif delta > 0.0: max_speed = c[8]
      else: max_speed = c[8]
        
      if not jumping:
        if character.random.random() < c[11]: # Attack
          for competitor in character.level.characters:
            if not competitor is character:
              px, py, pz = character.transform_point(competitor.x, competitor.y, competitor.z, competitor.parent)
              if (-5.0 < pz < 0.0) and (-2.0 < py < 2.0) and (-5.0 < px < 5.0):
                for i in range(c[10]):
                  fx, fy, fz = character.transform_point(competitor.x, competitor.y, competitor.z, competitor.parent)
                  
                  mouse_x = min(1.0, max(-1.0, c[0] * fx))
                  mouse_y = min(max_speed, max(-max_speed, c[1] * abs(fx) + c[2] * fy + c[3] * fz + c[4] * no_ground))
                  yield slune_character.Action(slune_character.ACTION_WAIT, mouse_x, mouse_y)
                break
              
        else: # Dodge
          for competitor in character.level.characters:
            if not competitor is character:
              px, py, pz = competitor.transform_point(character.x, character.y, character.z, character.parent)
              if (-5.0 < pz < 0.0) and (-2.0 < py < 2.0) and (-5.0 < px < 5.0):
                h = character.random.random()
                if h < 0.3:
                  for i in range(10): yield slune_character.Action(slune_character.ACTION_WAIT, -1.0, mouse_y)
                elif h < 0.6:
                  for i in range(10): yield slune_character.Action(slune_character.ACTION_WAIT,  1.0, mouse_y)
                else:
                  jumping = 1
                  
                  yield slune_character.Action(slune_character.ACTION_JUMP, mouse_x, mouse_y)
                  for i in range(2): yield slune_character.Action(slune_character.ACTION_ROTDOWN, mouse_x, mouse_y)
                break
